library ac_dart.ac_jwt.qsh;

import 'package:path/path.dart' as _path;
import 'package:quiver/collection.dart';
import 'package:collection/algorithms.dart';
import 'package:crypto/crypto.dart';
import 'precondition.dart';
import 'package:logging/logging.dart';

Logger _log = new Logger('ac_dart.ac_jwt.qsh');

const _CANONICAL_REQUEST_SEPARATOR = '&';
final _ENCODED_CANONICAL_REQUEST_SEPARATOR =
    Uri.encodeComponent(_CANONICAL_REQUEST_SEPARATOR);

/// Creates a query string hash from a [nonCanonicalUri], http method and
/// optional request [contextPath]
String createQsh(Uri nonCanonicalUri, String method, [String contextPath]) {
  final String canonicalUri =
      new CanonicalUri(nonCanonicalUri, method, contextPath).canonicalUri;
  return _qsh(canonicalUri);
}

/// Creates a query string hash from a relative [path] that excludes any
/// contextPath, http method and request [queryParameters]
String createQshFromPath(
    String path, String method, Multimap<String, String> queryParameters) {
  final String canonicalUri =
      new CanonicalUri.fromPath(path, method, queryParameters).canonicalUri;
  return _qsh(canonicalUri);
}

String _qsh(String canonicalUri) =>
    CryptoUtils.bytesToHex((new SHA256()..add(canonicalUri.codeUnits)).close());

/// Represents a uri in the canonnical form for creation of a query string hash
/// as defined by the Atlassian Connect documentation
class CanonicalUri {
  String _canonicalUri;
  String get canonicalUri => _canonicalUri;

  void set canonicalUri(String value) {
    _canonicalUri = value;
    _log.fine('canonical uri is $_canonicalUri');
  }

  CanonicalUri(Uri nonCanonicalUri, String method, [String contextPath]) {
    canonicalUri = _canonicalise(
        checkNotNull(nonCanonicalUri, 'nonCanonicalUri'),
        checkNotNull(method, 'method'),
        contextPath);
  }

  /// [path] assumed not to have context
  CanonicalUri.fromPath(
      String path, String method, Multimap<String, String> queryParameters) {
    canonicalUri = _createCanonicalUri(
        method, path, _canonicaliseParams(_encodeParams(queryParameters)));
  }

  static String _canonicalise(
      Uri nonCanonicalUri, String method, String contextPath) {
    final String canonicalPath =
        _canonicalisePath(nonCanonicalUri.path, contextPath);
    final String canonicalParams =
        _canonicaliseParams(_splitQueryParams(nonCanonicalUri.query));

    return _createCanonicalUri(method, canonicalPath, canonicalParams);
  }

  static String _createCanonicalUri(
      String method, String canonicalPath, String canonicalParams) {
    String addSlashIfRequired(String p) {
      return p.startsWith(_path.separator) ? p : '${_path.separator}$p';
    }

    final String path = addSlashIfRequired(canonicalPath)
        // If the separator is not URL encoded then the following URLs have the same query-string-hash:
        //   https://something.com/rest/api/2/project&a=b?x=y
        //   https://something.com/rest/api/2/project?a=b&x=y
        .replaceAll(
            _CANONICAL_REQUEST_SEPARATOR, _ENCODED_CANONICAL_REQUEST_SEPARATOR);

    // TODO: not sure why we are ending up with extra %25
    return '$method&$path&$canonicalParams'.replaceAll('%25', '%');
  }

  static String _canonicalisePath(String path, String contextPath) {
    String stripContextPath() {
      // TODO: starting to regret avoiding a regex approach

      List<String> skipContext(
          List<String> pathSegs, List<String> contextSegs) {
        if (pathSegs.isEmpty || contextSegs.isEmpty) {
          return pathSegs;
        }

        if (pathSegs.first == _path.separator) {
          return skipContext(pathSegs.sublist(1), contextSegs);
        }

        if (contextSegs.first == _path.separator) {
          return skipContext(pathSegs, contextSegs.sublist(1));
        }

        if (pathSegs.first == contextSegs.first) {
          return skipContext(pathSegs.sublist(1), contextSegs.sublist(1));
        }

        throw new ArgumentError(
            'path ($path) does not contain complete context' +
                ' path ($contextPath)');
      }

      final segs = skipContext(_path.split(path), _path.split(contextPath));
      return '${_path.separator}${_path.joinAll(segs)}';
    }

    return (contextPath == null || contextPath.trim().isEmpty)
        ? path
        : stripContextPath();
  }

  static String _canonicaliseParams(Multimap<String, String> queryParameters) {
    if (queryParameters.isEmpty) {
      return '';
    }

    List<String> paramList = [];
    queryParameters.toMap().forEach((k, v) {
      if (k != 'jwt') {
        paramList.add('$k=${_sort(v).join(",")}');
      }
    });
    return '${_sort(paramList).join(_CANONICAL_REQUEST_SEPARATOR)}';
  }

  static Iterable<String> _sort(Iterable<String> i) {
    final l = i.toList();
    insertionSort(l);
    return l;
  }

  // Uri.queryParameters does not support repeat parameters
  // Note: uses encodeComponent rather than encodeQueryComponent as spaces need
  // to be converted to %20 not +
  static Multimap<String, String> _splitQueryParams(String query) {
    return query
        .split(_CANONICAL_REQUEST_SEPARATOR)
        .fold(new Multimap(), (map, element) {
      int index = element.indexOf("=");
      if (index == -1) {
        if (element != "") {
          map.add(Uri.encodeComponent(element), "");
        }
      } else if (index != 0) {
        var key = element.substring(0, index);
        var value = element.substring(index + 1);
        map.add(Uri.encodeComponent(key), Uri.encodeComponent(value));
      }
      return map;
    });
  }

  static Multimap<String, String> _encodeParams(
      Multimap<String, String> params) {
    Multimap<String, String> encodedParams = new Multimap();

    params.forEach((k, v) {
      encodedParams.add(Uri.encodeComponent(k), Uri.encodeComponent(v));
    });

    return encodedParams;
  }
}
